package gs.config

import cats.effect.Sync
import cats.syntax.all.*
import gs.config.audit.ConfigManifest
import gs.config.audit.ConfigQueryResult
import gs.config.source.ConfigSource
import gs.config.source.EnvironmentConfigSource
import gs.config.source.MemoryConfigSource

final class AuditedConfiguration[F[_]: Sync](
  val sources: List[ConfigSource[F]],
  val manifest: ConfigManifest[F]
) extends BaseConfiguration[F]:
  import AuditedConfiguration.Acc

  override def getValue[A: Configurable](
    key: ConfigKey[A]
  ): F[Either[ConfigError, A]] =
    findAndParse(key)
      .map { acc =>
        acc -> (
          acc.result match
            case None         => handleMissingValue(key)
            case Some(result) => parse(key, result, acc.lastAttemptedSource)
        )
      }
      .flatMap { case (acc, result) => audit(key, result, acc).as(result) }

  private def findAndParse[A: Configurable](
    key: ConfigKey[A]
  ): F[Acc] =
    sources.foldLeft(Acc.empty[F]) {
      (
        acc,
        src
      ) =>
        acc.flatMap { data =>
          if data.hasResult then Sync[F].pure(data)
          else
            src.getValue(key).map {
              case None    => data.appendSource(src.name)
              case Some(v) => data.appendSource(src.name).withResult(v)
            }
        }
    }

  private def audit(
    key: ConfigKey[?],
    result: Either[ConfigError, ?],
    acc: Acc
  ): F[Unit] =
    result match
      case Left(error) =>
        manifest.record(
          name = key.name,
          queryResult = ConfigQueryResult.Failure(
            sources = acc.sources.toList,
            error = error
          )
        )
      case Right(_) =>
        manifest.record(
          name = key.name,
          queryResult = ConfigQueryResult.Success(
            source = acc.lastAttemptedSource,
            rawValue = acc.result.getOrElse("")
          )
        )

object AuditedConfiguration:

  def forSources[F[_]: Sync](sources: List[ConfigSource[F]])
    : F[Configuration[F]] =
    ConfigManifest
      .standard[F]
      .map(manifest =>
        new AuditedConfiguration[F](
          sources = sources,
          manifest = manifest
        )
      )

  def builder[F[_]: Sync]: Builder[F] = Builder[F](Vector.empty)

  case class Builder[F[_]: Sync](sources: Vector[ConfigSource[F]]):

    def withEnvironmentSource(): Builder[F] =
      copy(sources = this.sources.appended(new EnvironmentConfigSource[F]))

    def withMemorySource(configs: Map[String, String]): Builder[F] =
      copy(sources = this.sources.appended(new MemoryConfigSource[F](configs)))

    def withSource(source: ConfigSource[F]): Builder[F] =
      copy(sources = this.sources.appended(source))

    def build(): F[Configuration[F]] =
      ConfigManifest
        .standard[F]
        .map(manifest =>
          new AuditedConfiguration[F](
            sources = sources.toList,
            manifest = manifest
          )
        )

  /** Internal class for accumulating a result across sources.
    *
    * @param sources
    *   The seen sources.
    * @param result
    *   The result, or `None` if no result has been found.
    */
  private case class Acc(
    sources: Vector[String],
    result: Option[String]
  ):
    def hasResult: Boolean = result.isDefined

    def lastAttemptedSource: String = sources.lastOption.getOrElse("")

    def withResult(r: String): Acc =
      if hasResult then this else this.copy(sources, Some(r))

    def appendSource(s: String): Acc =
      if hasResult then this else this.copy(sources.appended(s), result)

  private object Acc:
    def empty[F[_]: Sync]: F[Acc] = Sync[F].pure(Acc(Vector.empty, None))

end AuditedConfiguration
